// =====================================================================
//  prefsum.c
// =====================================================================
 
//  Author:         (c) 2013 Camil Demetrescu
//  License:        See the end of this file for license information
//  Created:        May 26, 2013
 
//  Last changed:   $Date: 2013/05/26 13:01:07 $
//  Changed by:     $Author: demetres $
//  Revision:       $Revision: 1.00 $


#include <stdio.h>
#include <stdlib.h>

#include "clut.h"

#define ARRAY_SIZE      1048576
#define PROG_NAME       "prefsum.cl"
#define KERNEL_NAME     "prefsum"
#define LOCAL_SIZE      512


// ---------------------------------------------------------------------
// prefsum
// ---------------------------------------------------------------------
// sequential CPU version
void prefsum(int* in, int* out, int n, double* t) {

    int i;

    double start = clut_get_real_time();

    out[0] = in[0];
    for (i=1; i<n; i++) out[i] = out[i-1] + in[i];
    
    *t = clut_get_real_time() - start;
}


// ---------------------------------------------------------------------
// prefsum_cl
// ---------------------------------------------------------------------
// data-parallel GPU version
void prefsum_cl(int* in, int* out, int n, 
                clut_device* dev, double* td) {

    int       err;      // error code
    cl_kernel kernel;   // execution kernel
    cl_mem    dv1;      // input/output array on device
    cl_mem    dv2;      // input/output array on device
    cl_event  evt;
    int       offset;

    // create the compute kernel
    kernel = clCreateKernel(dev->program, KERNEL_NAME, &err);
    clut_check_err(err, "failed to create kernel");

    // allocate first array on device as a copy of array on host
    dv1 = clCreateBuffer(dev->context, 
                        CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR, 
                        n*sizeof(int), in, NULL);
    if (!dv1) clut_panic("failed to allocate array on device memory");

    // allocate second array on device
    dv2 = clCreateBuffer(dev->context, 
                        CL_MEM_READ_WRITE, 
                        n*sizeof(int), NULL, NULL);
    if (!dv2) clut_panic("failed to allocate array on device memory");

    size_t local_dim  = LOCAL_SIZE;
    size_t global_dim = n;
    global_dim = ((global_dim+LOCAL_SIZE-1)/LOCAL_SIZE)*LOCAL_SIZE; // round up

    *td = 0;

    // perform log_2(n) parallel rounds
    for (offset = 1; offset < n; offset <<= 1) {
        
        // set the arguments to our compute kernel
        err  = clSetKernelArg(kernel, 0, sizeof(cl_mem), &dv1);
        err |= clSetKernelArg(kernel, 1, sizeof(cl_mem), &dv2);
        err |= clSetKernelArg(kernel, 2, sizeof(int), &offset);
        err |= clSetKernelArg(kernel, 3, sizeof(int), &n);
        clut_check_err(err, "failed to set kernel arguments");

        // execute the kernel over NDRange of decreasing size
        err = clEnqueueNDRangeKernel(dev->queue, kernel, 1, 
                                     NULL, &global_dim, &local_dim, 0, NULL, &evt);
        clut_check_err(err, "failed to execute kernel");

        // wait for round termination
        clFinish(dev->queue);

        // accumulate time required by the kernel execution
        *td += clut_get_duration(evt);

        // swap buffers
        cl_mem temp = dv1;
        dv1 = dv2;
        dv2 = temp;
    }

    // copy result from device to host
    err = clEnqueueReadBuffer(dev->queue, dv1, CL_TRUE, 0, 
                              n*sizeof(int), out, 0, NULL, NULL);
    clut_check_err(err, "failed to read output result");

    clReleaseMemObject(dv1);
    clReleaseMemObject(dv2);
    clReleaseKernel(kernel);
}


// ---------------------------------------------------------------------
// equal_arrays
// ---------------------------------------------------------------------
int equal_arrays(int* a, int* b, int n){
    int i;
    for (i=0; i<n; i++)
        if (a[i] != b[i]) return 0;
    return 1;
}


// ---------------------------------------------------------------------
// run_test
// ---------------------------------------------------------------------
void run_test(int* v, int n, clut_device* dev, char* msg) {

    double t, td;
    
    int* out1 = malloc(n*sizeof(int));
    int* out2 = malloc(n*sizeof(int));
    if (out1 == NULL || out2 == NULL) 
        clut_panic("can't allocate output arrays on host");

    // compute prefix sums on both host and device
    prefsum(v, out1, n, &t);
    prefsum_cl(v, out2, n, dev, &td);
    
    // validate our results
    printf("%s\n", msg);
    printf("    CPU took %.3f msec\n", t*1E03);
    printf("    Device took %.3f msec\n", td*1E03);
    printf("    Correctness test: %s\n", equal_arrays(out1, out2, n) ? 
               "PASSED" : "FAILED");

    free(out1);
    free(out2);
}


// ---------------------------------------------------------------------
// main
// ---------------------------------------------------------------------
int main(int argc, char** argv) {

    clut_device dev;                // device structure
    int         n = ARRAY_SIZE;     // array size
    int*        v;                  // input array on host
    int         i;

    // override default array size
    if (argc > 1) n = atoi(argv[1]);

    // open device
    clut_open_device(&dev, PROG_NAME);

    printf("Prefix sum on array of size %d\n", n);
    printf("Work group size: %d\n", LOCAL_SIZE);
    clut_print_device_info(&dev);

    // allocate array on host
    v = malloc(n*sizeof(int));
    if (v == NULL) clut_panic("failed to allocate array on host");
    
    // make input array
    for (i = 0; i < n; i++) v[i] = 1;
    run_test(v, n, &dev,"Test:");

    // cleanup
    free(v);
    clut_close_device(&dev);
    
    return 0;
}


// Copyright (C) 2013 Camil Demetrescu
  
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
  
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
  
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
// USA
